from pathlib import Path

import click

from ..commands import *
from ..utils.config_manager import get_config_data

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


@click.group(context_settings=CONTEXT_SETTINGS)
def cli():
    global config_data
    config_data = get_config_data()


@cli.command("create", short_help="create a task")
@click.argument("filename", type=str)
@click.option(
    "-p",
    "--path",
    "base_folder",
    default=None,
    type=click.Path(exists=True,
                    path_type=Path,
                    writable=True,
                    file_okay=False),
    help="path to the folder where you want to create the source code file")
@click.option("-f",
              "--force",
              is_flag=True,
              help="overwrite existing files of task")
def create(filename, base_folder, force):
    """
    \b
    Create a task to add and run testcases

    \b
    Args:

    \b
    FILENAME of the file to be created with file extension
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    create_manage(filename, base_folder, force, config_data)


@cli.command("fetch", short_help="retrieve testcase data from online judge")
@click.option(
    "-p",
    "--path",
    "base_folder",
    default=None,
    type=click.Path(exists=True,
                    path_type=Path,
                    writable=True,
                    file_okay=False),
    help="path to the folder where you want to create the source code file")
@click.option("-f",
              "--force",
              is_flag=True,
              help="overwrite existing files of task")
def fetch(base_folder, force):
    """
    \b
    Retrieve testcase data from online judge,
    make the source code file,
    and copy boiler plate code if specified in config file
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    fetch_manage(base_folder, force, config_data)


@cli.command("run", short_help="run code against testcases")
@click.argument("filename", type=str)
@click.option("-p",
              "--path",
              "base_folder",
              default=None,
              type=click.Path(exists=True,
                              path_type=Path,
                              writable=True,
                              file_okay=False),
              help="path to the folder which contains the souce code")
@click.option("-t",
              "--tc",
              default=0,
              show_default=True,
              help="run specific testcase (0 for all)")
@click.option(
    "-i",
    "--interactive",
    is_flag=True,
    help="run program in interactive mode(stdin, stdout, stderr are used)")
def run(filename, base_folder, tc, interactive):
    """
    \b
    Compile (if applied) and run source code on saved testcases
    using commands specified in the config file

    \b
    Args:
    
    \b
    FILENAME of the file to run with file extension
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    run_manage(filename, base_folder, config_data, tc, interactive)


@cli.command("compile", short_help="compile source code")
@click.argument("filename", type=str)
@click.option("-p",
              "--path",
              "base_folder",
              default=None,
              type=click.Path(exists=True,
                              path_type=Path,
                              writable=True,
                              file_okay=False),
              help="path to the folder which contains the souce code")
def compile(filename, base_folder):
    """
    \b
    Compile (if applied) the source code 
    using commands specified in the config file

    \b
    Args:
    
    \b
    FILENAME of the file to run with file extension
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    compile_manage(filename, base_folder, config_data)


@cli.command("addtc", short_help="add a tc to be used in future runs of code")
@click.argument("filename", type=str)
@click.argument("input-path",
                type=click.Path(exists=True, path_type=Path, dir_okay=False))
@click.argument("output-path",
                type=click.Path(exists=True, path_type=Path, dir_okay=False))
@click.option("-p",
              "--path",
              "base_folder",
              default=None,
              type=click.Path(
                  exists=True,
                  path_type=Path,
                  writable=True,
                  file_okay=False,
              ),
              help="path to the folder which contains the souce code")
def addtc(filename, input_path, output_path, base_folder):
    """
    \b
    Add a testcase to be used in future runs of code

    \b
    Args:
    
    \b
    FILENAME of the source code file with file extension 
    INPUT_PATH: path of file containing input of the testcase
    OUTPUT_PATH: path of file containing output of the testcase
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    addtc_manage(filename, input_path, output_path, base_folder)


@cli.command("test", short_help="brute force testing")
@click.argument("filename", type=str)
@click.argument("testcase-generator-path",
                type=click.Path(exists=True, path_type=Path, dir_okay=False))
@click.option("-n",
              "--number-of-runs",
              default=10000,
              show_default=True,
              help="no of randomly generated testcases")
@click.option("-p",
              "--path",
              "base_folder",
              default=None,
              type=click.Path(
                  exists=True,
                  path_type=Path,
                  writable=True,
                  file_okay=False,
              ),
              help="path to the folder which contains the souce code")
def test(filename, number_of_runs, testcase_generator_path, base_folder):
    """
    \b
    Generate random testcases and run your code against them.
    Create a function in your source code which validates 
    your answer if the validation fails print your answer to
    std output.
    \b
    Args:
    
    \b
    FILENAME of the source code file with file extension 
    TESTCASE_GENERATOR_PATH: path of test case generator file
    """
    if base_folder is None:
        base_folder = Path(config_data["default_base_folder"])
    test_manage(filename, number_of_runs, testcase_generator_path, base_folder,
                config_data)


@cli.command("config",
             short_help="get location of config file or reset config file")
@click.option("--reset",
              "-r",
              is_flag=True,
              help="reset config file to default values")
def config(reset):
    """
    \b
    Locate config file or reset config to default values with --reset option
    """
    config_manage(reset)



import re
import subprocess
import sys
from pathlib import Path

import click

from .config_manager import get_config_path


# checks if file is empty (only contains white spaces)
def is_file_empty(file_path):
    content = open(file_path, 'r').read()
    if re.search(r'^\s*$', content):
        return True
    return False


def check_diff(file_path1, file_path2):
    file1_txt = (read_from_file(file_path1)).split()
    file2_txt = (read_from_file(file_path2)).split()
    if len(file1_txt) != len(file2_txt):
        return False
    for i in range(len(file1_txt)):
        if file1_txt[i] != file2_txt[i]:
            return False
    return True


def write_to_file(file_path, text):
    with open(file_path, "w") as file:
        file.write(text)


def read_from_file(file_path):
    file = open(file_path, "r")
    text = "".join(file.readlines())
    file.close()
    return text


def print_file(file_path, is_error=0):
    """
        is_error = 0 -> normal text
        is_error = 1 -> warning text
        is_error = 2 -> error text
    """
    if not file_path.exists():
        click.secho(f"File at {file_path.resolve()} does not exist",
                    fg="red",
                    err=True)
        sys.exit(1)
    elif not file_path.is_file():
        click.secho(f"{file_path.resolve()} is not a file", fg="red", err=True)
        sys.exit(1)

    with open(file_path, 'r') as file:
        for line in file.readlines():
            if is_error == 2:
                click.secho(line, fg="red", err=True, nl=False)
            elif is_error == 1:
                click.secho(line, fg="yellow", err=True, nl=False)
            elif is_error == 0:
                click.secho(line, nl=False)
    click.echo("")


def open_source_code_in_editor(config_data, file_path):
    if "editor" in config_data:
        if config_data["editor"] is not None:
            click.secho(
                f"Opening {file_path.resolve()} with {config_data['editor']}",
                fg="cyan")
            try:
                subprocess.run([config_data['editor'], file_path])
            except OSError:
                click.secho(f"Command not found: {config_data['editor']}",
                            fg="red")
                click.secho(
                    f"Try changing the command in the config file (located at {get_config_path()})",
                    fg="cyan")
                sys.exit(1)



import re
import subprocess
import sys
from pathlib import Path

import click

from ...utils.config_manager import get_config_path
from ...utils.file_manager import (check_diff, is_file_empty, print_file,
                                   read_from_file, write_to_file)
from ...utils.folder_manager import Task, clear_folder


def cpp_compile(source_code_path, error_path, command):
    file_name_without_extension = source_code_path.stem
    exec_path = source_code_path.parent / file_name_without_extension
    command = command.split()
    command.extend([source_code_path, "-o", exec_path])
    try:
        compilation_data = subprocess.run(command,
                                          capture_output=True,
                                          text=True)
    except OSError:
        click.secho(f"Command not found: {' '.join(map(str, command))}",
                    fg="red",
                    err=True)
        click.secho(
            f"Try changing the command in the config file (located at {get_config_path()})",
            fg="cyan")
        sys.exit(1)

    write_to_file(error_path, compilation_data.stderr)
    return [exec_path, compilation_data.returncode]


def cpp_run(exec_path,
            input_path=None,
            output_path=None,
            error_path=None,
            interactive=False):
    if not interactive:
        try:
            run_data = subprocess.run([exec_path.resolve()],
                                      input=read_from_file(input_path),
                                      capture_output=True,
                                      text=True)
        except Exception as e:
            click.secho(f"Unknown Error while running {exec_path.resolve()}",
                        fg="red",
                        err=True)
            sys.exit(1)
        write_to_file(error_path, run_data.stderr)
        write_to_file(output_path, run_data.stdout)
        return run_data.returncode
    else:
        subprocess.run([exec_path.resolve()])


def java_compile(source_code_path, error_path, command):
    file_name_without_extension = source_code_path.stem
    exec_path = source_code_path.parent / file_name_without_extension
    command = command.split()
    command.extend([source_code_path])
    try:
        compilation_data = subprocess.run(command,
                                          capture_output=True,
                                          text=True)
    except OSError:
        click.secho(f"Command not found: {' '.join(map(str, command))}",
                    fg="red",
                    err=True)
        click.secho(
            f"Try changing the command in the config file (located at {get_config_path()})",
            fg="cyan")
        sys.exit(1)
    write_to_file(error_path, compilation_data.stderr)

    return [exec_path, compilation_data.returncode]


def java_run(exec_path,
             input_path=None,
             output_path=None,
             error_path=None,
             interactive=False):
    if not interactive:
        try:
            run_data = subprocess.run(["java", exec_path],
                                      input=read_from_file(input_path),
                                      capture_output=True,
                                      text=True)
        except Exception as e:
            click.secho(
                f"Unknown Error while running {'java' + str(exec_path.resolve())}",
                fg="red",
                err=True)
            sys.exit(1)
        write_to_file(error_path, run_data.stderr)
        write_to_file(output_path, run_data.stdout)
        return run_data.returncode
    else:
        subprocess.run(["java", exec_path])


# exec path incase of interpreted language is source code path
def py_run(source_code_path,
           command,
           input_path=None,
           output_path=None,
           error_path=None,
           interactive=False):
    command = command.split()
    command.extend([source_code_path])
    if not interactive:
        try:
            run_data = subprocess.run(command,
                                      input=read_from_file(input_path),
                                      capture_output=True,
                                      text=True)
        except OSError:
            click.secho(f"Command not found: {' '.join(map(str, command))}",
                        fg="red",
                        err=True)
            click.secho(
                f"Try changing the command in the config file (located at {get_config_path()})",
                fg="cyan")
            sys.exit(1)
        except Exception as e:
            click.secho(
                f"Unknown Error while running {' '.join(map(str, command))}",
                fg="red",
                err=True)
            sys.exit(1)
        write_to_file(error_path, run_data.stderr)
        write_to_file(output_path, run_data.stdout)
        return run_data.returncode
    else:
        subprocess.run(command)


is_interpreted = {
    "py": True,
    "java": False,
    "cpp": False,
}

run_func = {
    "py": py_run,
    "java": java_run,
    "cpp": cpp_run,
}

compile_func = {
    "java": java_compile,
    "cpp": cpp_compile,
}


def compile_source_code(source_code_path,
                        compilation_error_path,
                        extension,
                        config_data,
                        gen_compile=False):
    if not gen_compile:
        click.secho(f"Compiling the source code with command:", fg="cyan")
    else:
        click.secho(f"Compiling the generator code with command:", fg="cyan")
    click.secho(config_data["language"][extension]["command"] + "\n")
    exec_path, compile_returncode = compile_func[extension](
        source_code_path, compilation_error_path,
        config_data["language"][extension]["command"])
    if compile_returncode != 0:
        click.secho(f"Compilation Error:\n", fg="red")
        print_file(compilation_error_path, 2)
        sys.exit()
    else:
        click.secho(f"Compiled Successfully\n", fg="green")
        if not is_file_empty(compilation_error_path):
            click.secho(f"Compilation Warning:\n", fg="cyan")
            print_file(compilation_error_path, 1)
    return exec_path


def judge(task, filename_without_extension, extension, base_folder,
          config_data, tc, interactive):
    clear_folder(task.last_run_folder)

    source_code_path = (base_folder /
                        filename_without_extension).with_suffix("." +
                                                                extension)
    compilation_error_path = task.last_run_folder / "compilation_error.txt"

    exec_path = Path()
    if not is_interpreted[extension]:
        exec_path = compile_source_code(source_code_path,
                                        compilation_error_path, extension,
                                        config_data)
    else:
        click.secho(f"Running the source code with command:", fg="cyan")
        click.secho(config_data["language"][extension]["command"] + "\n")

    if not interactive:
        tc_list = []
        if tc == 0:
            tc_list = task.get_tc_list()
        else:
            tc_list.append(tc)

        if tc_list == []:
            click.secho(f"No testcase to run!", fg="red")
            sys.exit(1)

        for num in tc_list:
            in_path = task.tc_folder / f"in{num}.txt"
            ans_path = task.tc_folder / f"ans{num}.txt"
            if not in_path.is_file() and not ans_path.is_file():
                click.secho(f"Skipped #{num}\n", fg="yellow")
                click.secho(
                    f"Input file and Answer file both are not present\n",
                    fg="red",
                    err=True)
                continue
            elif not in_path.is_file():
                click.secho(f"Skipped #{num}\n", fg="yellow")
                click.secho(f"Input file is not present\n", fg="red", err=True)
                continue
            elif not ans_path.is_file():
                click.secho(f"Skipped #{num}\n", fg="yellow")
                click.secho(f"Answer file is not present\n",
                            fg="red",
                            err=True)
                continue
            std_output_path = task.last_run_folder / f"output{num}.txt"
            std_error_path = task.last_run_folder / f"error{num}.txt"

            click.secho(f"Running TC #{num}\n", fg="cyan")

            run_returncode = 0
            if is_interpreted[extension]:
                run_returncode = run_func[extension](
                    source_code_path,
                    config_data["language"][extension]["command"],
                    in_path,
                    std_output_path,
                    std_error_path,
                )
            else:
                run_returncode = run_func[extension](exec_path, in_path,
                                                     std_output_path,
                                                     std_error_path)

            if run_returncode != 0:
                click.secho(f"Rumtime Error #{num}\n", fg="red")
                print_file(std_error_path, 2)
                continue

            if (check_diff(std_output_path, ans_path)):
                click.secho(f"Accepted #{num}\n", fg="green")

                if not is_file_empty(std_error_path):
                    click.secho(f"Standard Error: ", fg="cyan")
                    print_file(std_error_path)
            else:
                click.secho(f"Wrong Answer #{num}\n", fg="red")

                click.secho(f"Test Case: ", fg="cyan")
                print_file(in_path)

                click.secho(f"Correct Answer: ", fg="cyan")
                print_file(ans_path)

                click.secho(f"Standard Output: ", fg="cyan")
                print_file(std_output_path)

                if not is_file_empty(std_error_path):
                    click.secho(f"Standard Error: ", fg="cyan")
                    print_file(std_error_path)
    else:
        click.secho("Enter the input:", fg="cyan")
        if is_interpreted[extension]:
            run_func[extension](source_code_path,
                                config_data["language"][extension]["command"],
                                interactive=True)
        else:
            run_func[extension](exec_path, interactive=True)


def manage(filename, base_folder, config_data, tc, interactive):
    file_path = base_folder / filename
    extension = file_path.suffix[1:]
    filename_without_extension = file_path.stem
    if extension not in ["py", "java", "cpp"]:
        click.secho(
            f"Language {extension} is not supported (python(py), java, c++(cpp) are only supported)",
            err=True,
            fg="red")
        sys.exit(1)
    else:
        task = Task(base_folder, filename_without_extension, extension)
        task_status = task.task_exists()
        if task_status not in [1, 3]:
            click.secho("Source Code does not exist", err=True, fg="red")
            sys.exit(1)
        else:
            if task_status == 1:
                task.create_task(create_source_code=False)
            judge(task, filename_without_extension, extension, base_folder,
                  config_data, tc, interactive)



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="cppt",
    version='1.2.0',
    author="Vishal Agrawal",
    author_email="vishalagrawalva22@gmail.com",
    description=
    "Command line tool to automate your competitive programming workflow without cluttering your workspace with testcase data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/vishalagrawal22/CPPT",
    py_modules=['cppt.bin.cppt'],
    include_package_data=True,
    install_requires=[
        'click',
        'pyyaml',
    ],
    packages=[
        "cppt.utils", "cppt.commands.create", "cppt.commands.fetch",
        "cppt.commands.addtc", "cppt.commands.run", "cppt.commands.config",
        "cppt.commands.compile", "cppt.commands.test", "cppt.commands"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: MIT License",
    ],
    python_requires='>=3.6',
    entry_points={
        'console_scripts': [
            'cppt = cppt.bin.cppt:cli',
        ],
    },
)
